/**
 * @file lv_demo_widgets.c
 *
 */

/*********************
 *      INCLUDES
 *********************/
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"
#include "../../lv_examples.h"
#include "lv_demo_widgets.h"
#include "fsl_common.h"

/*********************
 *      DEFINES
 *********************/

/**********************
 *      TYPEDEFS
 **********************/

/**********************
 *  STATIC PROTOTYPES
 **********************/
// We have N canvases; we unhide one and are allowed to plot into the others, rotating
// If this number is 2, it's the same as double buffering
// I've tried soooo many ways to make this fast, but it always seems to take 27ms to do one round of lv_task_handler
// I tried one canvas, using lv_canvas_set_buffer to switch between buffers, both in an lv_task, and in the main AppTask
// I tried one canvas, with an lv_task that draws directly into it and then invalidates it
// No matter what it ALWAYS takes 27ms
#define NCANVAS_MEM_BUFFERS 4 // This number doesn't even seem to matter w.r.t. timing
AT_NONCACHEABLE_SECTION(static lv_color_t s_canvasMem[NCANVAS_MEM_BUFFERS][LV_HOR_RES_MAX][LV_VER_RES_MAX]);
static uint32_t canvasMemDrawingIdx = 0;
static uint32_t canvasMemDisplayingIdx = 0;
static uint32_t canvasMemLastDisplayedIdx = 0;
SemaphoreHandle_t canvasQueue = 0;
lv_obj_t *canvas[NCANVAS_MEM_BUFFERS] = {0};

/**********************
 *  STATIC VARIABLES
 **********************/
static lv_style_t style_box;
/**********************
 *      MACROS
 **********************/
/**********************
 *   GLOBAL FUNCTIONS
 **********************/

#include "bresenham.h"

static void create_meas_plot_data(void);
void update_meas_plot_data(lv_task_t * t);
TaskHandle_t plotTaskHandle = 0;

// Hacky timing stuff
static uint32_t loopTimes[256] = {0};
static uint32_t linedrawTimes[256] = {0};
static uint32_t loopTimeIdx = 0;
static uint32_t linedrawTimeIdx = 0;
static uint32_t drawQ[1024][3] = {0};
static uint32_t drawQIdx = 0;
static uint32_t dispQ[1024][3] = {0};
static uint32_t dispQIdx = 0;

void PlotTask(void *p)
{
    while (1) {
        // Hacky timing stuff
        drawQ[drawQIdx][0] = xTaskGetTickCount();
        drawQ[drawQIdx][1] = uxSemaphoreGetCount(canvasQueue);
        drawQ[drawQIdx][2] = canvasMemDrawingIdx;
        drawQIdx = (drawQIdx + 1) % 1024;

        xSemaphoreTake(canvasQueue, portMAX_DELAY);

        // Hacky timing stuff
        linedrawTimes[linedrawTimeIdx] = xTaskGetTickCount();
        linedrawTimeIdx = (linedrawTimeIdx + 1) % 256;

        create_meas_plot_data();
        canvasMemDrawingIdx = (canvasMemDrawingIdx + 1) % NCANVAS_MEM_BUFFERS;
    }
}

void lv_demo_widgets(void)
{
    lv_style_init(&style_box);
    lv_style_set_value_align(&style_box, LV_STATE_DEFAULT, LV_ALIGN_OUT_TOP_LEFT);
    lv_style_set_value_ofs_y(&style_box, LV_STATE_DEFAULT, LV_DPX(0));
    lv_style_set_margin_top(&style_box, LV_STATE_DEFAULT, LV_DPX(0));

    memset(s_canvasMem, 0, sizeof(s_canvasMem));
    // Create N canvases; we are allowed to plot in all but the currently displaying one
    // We hide these by default and only make visible the displaying one later
    for (uint32_t i = 0; i < NCANVAS_MEM_BUFFERS; i++) {
        canvas[i] = lv_canvas_create(lv_scr_act(), NULL);
        lv_canvas_set_buffer(canvas[i], s_canvasMem[i], LV_HOR_RES_MAX, LV_VER_RES_MAX, LV_IMG_CF_TRUE_COLOR);
        lv_obj_set_hidden(canvas[i], true);
    }

    canvasQueue = xSemaphoreCreateCounting(NCANVAS_MEM_BUFFERS-1, NCANVAS_MEM_BUFFERS-1);
    xTaskCreate(PlotTask, "plot", configMINIMAL_STACK_SIZE + 800, NULL, tskIDLE_PRIORITY + 2, &plotTaskHandle);
}


/**********************
 *   STATIC FUNCTIONS
 **********************/
#define DRAW_NONSTRAIGHT_LINES 0 // this doesn't seem to matter much for performance
#include <stdlib.h>
// Drawing 640 lines takes around 4ms
static void create_meas_plot_data(void)
{
    lv_color_t colorS = {
            .ch.blue = 0x1F,
            .ch.green = 0,
            .ch.red = 0,
    };
    uint16_t color = colorS.full;
    memset(s_canvasMem[canvasMemDrawingIdx], 0xFF, sizeof(s_canvasMem[canvasMemDrawingIdx]));

    uint16_t y0 = 0;
    int16_t yd = 0;
    int16_t y = 0;
#if DRAW_NONSTRAIGHT_LINES == 1
    for (uint16_t i = 1; i < LV_HOR_RES_MAX; i++)
#else
    for (uint16_t i = 0; i < LV_HOR_RES_MAX; i++)
#endif
    {
        yd = rand() % 15 + 15;
        if (rand() % 2 == 0) {
            yd = -yd;
        }
        y += yd;
        if ((y >= LV_VER_RES_MAX) || (y < 0))
        {
            yd = -yd;
            y = y + yd + yd;
        }
#if DRAW_NONSTRAIGHT_LINES == 1
        mmDrawLineFaster(s_canvasMem[canvasMemDrawingIdx], LV_HOR_RES_MAX, i-1, y0, i, y, color);
#else
        mmDrawLineFaster(s_canvasMem[canvasMemDrawingIdx], LV_HOR_RES_MAX, i, y0, i, y, color);
#endif
        y0 = y;
    }
}


void update_meas_plot_data(lv_task_t * t)
{
    // Hacky timing stuff
    dispQ[dispQIdx][0] = xTaskGetTickCount();
    dispQ[dispQIdx][1] = uxSemaphoreGetCount(canvasQueue);
    dispQ[dispQIdx][2] = canvasMemDisplayingIdx;
    dispQIdx = (dispQIdx + 1) % 1024;

    // If there isn't anything to display, don't do anything
    if (uxSemaphoreGetCount(canvasQueue) == NCANVAS_MEM_BUFFERS) {
        return;
    }

    // Hacky timing stuff
    loopTimes[loopTimeIdx] = xTaskGetTickCount();
    loopTimeIdx = (loopTimeIdx + 1) % 256;

    // Hide the last canvas we displayed
    lv_obj_set_hidden(canvas[canvasMemLastDisplayedIdx], true);
    canvasMemLastDisplayedIdx = canvasMemDisplayingIdx;

    // Show the current canvas
    lv_obj_set_hidden(canvas[canvasMemDisplayingIdx], false);
    canvasMemDisplayingIdx = (canvasMemDisplayingIdx + 1) % NCANVAS_MEM_BUFFERS;

    xSemaphoreGive(canvasQueue);
}

